# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import os
import errno
import logging


LOG_LEVELS = frozenset([
    'CRITICAL',
    'DEBUG',
    'ERROR',
    'FATAL',
    'INFO',
    'NOTSET',
    'WARN',
    'WARNING',
])
LOG_FORMAT = '%(asctime)s %(levelname)s: [%(name)s] %(message)s'


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


class NullLogger(logging.Logger):
    def __init__(self):
        logging.Logger.__init__(self, "null")
        self.propagate = 0
        self.handlers = [NullHandler()]


def get_log_level(s):
    try:
        return int(s)
    except ValueError:
        level_name = s.upper()
        if level_name in LOG_LEVELS:
            return getattr(logging, level_name)
        else:
            raise ValueError('%r is not a valid log level' % (s,))


def get_logger(name, level=None, f=None, mode='a'):
    if level is None:
        return NullLogger()
    logger = logging.getLogger(name)
    logger.setLevel(get_log_level(level))
    if isinstance(f, str):
        handler = logging.FileHandler(f, mode=mode)
    else:
        handler = logging.StreamHandler(f)
    handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(handler)
    return logger


def mkdir_p(*paths):
    for p in paths:
        try:
            os.makedirs(p)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

from configparser import ConfigParser

RELPATH = "biotracks.ini"
TOP_LEVEL = "TOP_LEVEL_INFO"
TRACKING = "TRACKING_DATA"
DEFAULT_NAME = "cmso_tracks"


def get_conf(conf_fn=None):
    conf = ConfigParser()
    if conf_fn:
        conf.read(conf_fn)
    else:
        for section in TOP_LEVEL, TRACKING:
            conf.add_section(section)
    conf.setdefault(TOP_LEVEL, {}).setdefault("name", DEFAULT_NAME)
    return conf



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import os
import datapackage as dp
try:
    from datapackage.mappers import convert_path  # datapackage version 0.x
except ImportError:
    from datapackage.pushpull import _convert_path as convert_path
import pandas as pd

from .utils import get_logger
from . import cmso


def push_to_pandas(directory, object_id_cmso, log_level=None):
    """Push a datapackage to a pandas storage.

    arguments:
    directory -- the datapackage directory containing the json file
    object_id_cmso -- the object id
    """
    logger = get_logger('push_to_pandas', level=log_level)
    descr = directory + os.sep + 'dp.json'
    storage = dp.push_datapackage(descriptor=descr, backend='pandas')
    objects = storage[convert_path("objects.csv", cmso.OBJECTS_TABLE)]
    links = storage[convert_path("links.csv", cmso.LINKS_TABLE)]
    objects.reset_index(inplace=True)

    # aggregation to construct tracks
    tracks_dict = {}
    list_ = []
    TRACK_ID = -1
    for link in links[cmso.LINK_ID].unique():
        tmp = links[links[cmso.LINK_ID] == link]
        rest = links[
            (links[cmso.LINK_ID] != link) & (~links[cmso.LINK_ID].isin(list_))
        ]
        ind = rest[object_id_cmso].isin(tmp[object_id_cmso])
        # no shared spots
        if not any(ind):
            # link not in dictionary
            if link not in [l for v in tracks_dict.values() for l in v]:
                TRACK_ID += 1
                tracks_dict[TRACK_ID] = [link]
        # shared spots
        if any(ind):
            for index, b in enumerate(ind):
                if b:
                    # link not in dictionary
                    if link not in [l for v in tracks_dict.values()
                                    for l in v]:
                        TRACK_ID += 1
                        tracks_dict[TRACK_ID] = []
                    tracks_dict[TRACK_ID].append(link)
                    tracks_dict[TRACK_ID].append(
                        rest.iloc[index][cmso.LINK_ID]
                    )
        list_.append(link)

    # get unique links and construct tracks dataframe
    tracks_dict_unique = {}
    for key, value in tracks_dict.items():
        unique_set = set(value)
        tracks_dict_unique[key] = unique_set
    logger.debug('Created %d tracks', len(tracks_dict_unique))
    tracks = pd.DataFrame()
    for key, value in tracks_dict_unique.items():
        for link in value:
            tracks = tracks.append([[key, link]], ignore_index=True)
    tracks.columns = [cmso.TRACK_ID, cmso.LINK_ID]
    return {'objects': objects, 'links': links, 'tracks': tracks}



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import os


def get_obj_dict(df, obj_id):
    d = {}
    for _, series in df.iterrows():
        k = series.pop(obj_id)
        assert k not in d
        d[k] = series.to_dict()
    return d


def get_link_dict(df, obj_id, link_id):
    d = {}
    for _, series in df.iterrows():
        d.setdefault(series[link_id], set()).add(series[obj_id])
    return d


def get_track_dict(df, link_id, track_id):
    return get_link_dict(df, link_id, track_id)  # same logic


THIS_DIR = os.path.dirname(os.path.abspath(__file__))
EXAMPLES_DIR = os.path.join(THIS_DIR, os.pardir, 'examples')
RELPATHS = {
    'ICY': ['example_2', 'track_processor_ICY.xls'],
    'CELLMIA': ['example_1', '0001_mode1_z000_f000_tracking.txt'],
    'TrackMate': ['example_1', 'FakeTracks.xml'],
    'CellProfiler': ['example_1', 'output', 'bloboverlap15_spots.csv'],
    'Mosaic': ['example_1', 'mosaic_tracks.csv'],
}



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import os

import datapackage
import pytest

from biotracks import createdp, readfile, config
from .common import EXAMPLES_DIR, RELPATHS


@pytest.fixture()
def data(tmpdir):
    def make_data(fmt):
        base_dir = os.path.join(EXAMPLES_DIR, fmt, *RELPATHS[fmt][:-1])
        exp_dp_dir = os.path.join(base_dir, 'dp')
        dp_fn = os.path.join(exp_dp_dir, 'dp.json')
        dp = datapackage.DataPackage(dp_fn)
        in_fn = os.path.join(base_dir, RELPATHS[fmt][-1])
        conf_fn = os.path.join(base_dir, config.RELPATH)
        conf = config.get_conf(conf_fn=conf_fn)
        reader = readfile.TracksReader(in_fn, conf=conf)
        reader.read()
        return {'reader': reader, 'dp': dp, 'dp_dir': str(tmpdir)}
    yield make_data
    tmpdir.remove(rec=True)


class TestCreatedp(object):

    def test_icy(self, data):
        self.__check_dps(data('ICY'))

    def test_cellmia(self, data):
        self.__check_dps(data('CELLMIA'))

    def test_cellprofiler(self, data):
        self.__check_dps(data('CellProfiler'))

    def test_trackmate(self, data):
        self.__check_dps(data('TrackMate'))

    def test_mosaic(self, data):
        self.__check_dps(data('Mosaic'))

    def __check_dps(self, d):
        dp = createdp.create(d['reader'], d['dp_dir'])
        assert dp.to_dict() == d['dp'].to_dict()
        d['reader'].conf[config.TOP_LEVEL]['name'] = 'CMSO_TRACKS'
        with pytest.raises(ValueError):
            createdp.create(d['reader'], d['dp_dir'])



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

"""\
Reader objects for tracking data formats.
"""

import os
import xml.etree.ElementTree as ET
from abc import ABCMeta, abstractmethod

import pandas as pd
import xlrd

from .utils import get_logger
from .validation import Validator
from . import cmso, config


class AbstractReader(metaclass=ABCMeta):
    """\
    Common interface for all tracking data readers.
    """

    def __init__(self, fname, conf=None, log_level=None):
        """\
        Set the input file name for the reader.

        The ``conf`` argument is optional in the general case, but it
        may be required for a specific reader.

        Actual reading is supposed to happen when the ``read`` method
        is called.
        """
        reader_name = self.__class__.__name__
        self.logger = get_logger(reader_name, level=log_level)
        self.fname = fname
        self.conf = conf or config.get_conf()
        self.log_level = log_level
        self.logger.info('%s Reading "%s"', reader_name, fname)
        self._objects = None
        self._links = None

    @abstractmethod
    def read(self):
        """\
        Read the input file. The concrete implementation of this
        method should populate the _objects and _links attributes with
        dataframes containing the corresponding tracking data.  If the
        specific format contains additional information, for instance
        on time and space units, this should be added to the ``conf``
        object (if not already present).
        """
        return

    @property
    def objects(self):
        """\
        Return the objects dataframe, calling ``read`` if necessary.
        """
        if self._objects is None:
            self.read()
        return self._objects

    @property
    def links(self):
        """\
        Return the links dataframe, calling ``read`` if necessary.
        """
        if self._links is None:
            self.read()
        return self._links


class BiotracksReader(AbstractReader):

    def read(self):
        dp = Validator(log_level=self.log_level).validate(self.fname)
        d = os.path.dirname(self.fname)
        path_map = dict((_.descriptor["name"], _.descriptor["path"])
                        for _ in dp.resources)
        self._objects = pd.read_csv(
            os.path.join(d, path_map[cmso.OBJECTS_TABLE])
        )
        self._links = pd.read_csv(
            os.path.join(d, path_map[cmso.LINKS_TABLE])
        )
        for k in cmso.SPACE_UNIT, cmso.TIME_UNIT:
            try:
                v = dp.descriptor[k]
            except KeyError:
                pass
            else:
                self.conf[config.TOP_LEVEL].setdefault(k, v)


class TrackMateReader(AbstractReader):

    def read(self):
        self.root = ET.parse(self.fname).getroot()
        self.logger.info('Reading a TrackMate XML file version %s',
                         self.root.attrib.get('version'))
        self.__model = self.root.find('Model')
        self.conf[config.TOP_LEVEL].setdefault(
            cmso.SPACE_UNIT, self.__model.attrib["spatialunits"]
        )
        self.conf[config.TOP_LEVEL].setdefault(
            cmso.TIME_UNIT, self.__model.attrib["timeunits"]
        )
        spots_dict = self.read_spots()
        self._objects = pd.DataFrame(
            [[k, v[0], v[1], v[2]] for k, v in spots_dict.items()],
            columns=[cmso.OBJECT_ID, cmso.FRAME_ID, cmso.X_COORD, cmso.Y_COORD]
        )
        ordered_edges_df = self.read_edges(spots_dict)
        self._links = self.read_links(ordered_edges_df)

    def read_spots(self):
        spots_dict = {}
        for child in self.__model:
            if child.tag == 'AllSpots':
                spots = child
                for spot_in_frame in spots.getchildren():
                    for spot in spot_in_frame.getchildren():
                        spot_id = int(spot.get('ID'))
                        frame = int(spot.get('FRAME'))
                        position_x = float(spot.get('POSITION_X'))
                        position_y = float(spot.get('POSITION_Y'))
                        spots_dict[spot_id] = [frame, position_x, position_y]
        self.logger.debug('Found %d unique spots', len(spots_dict))
        return spots_dict

    def read_edges(self, spots_dict):
        edges_dict = {}
        edge_id = 0
        for child in self.__model:
            if child.tag == 'AllTracks':
                tracks = child
                for track in tracks.getchildren():
                    track_id = int(track.get('TRACK_ID'))
                    for edge in track.getchildren():
                        spot_source_id = int(edge.get('SPOT_SOURCE_ID'))
                        spot_target_id = int(edge.get('SPOT_TARGET_ID'))
                        source_frame = spots_dict.get(spot_source_id)[0]
                        target_frame = spots_dict.get(spot_target_id)[0]
                        edges_dict[edge_id] = [
                            track_id,
                            spot_source_id,
                            spot_target_id,
                            source_frame,
                            target_frame
                        ]
                        edge_id += 1
        self.logger.debug('Found %d unique edges', len(edges_dict))

        # dictionary into pandas dataframe
        edges_df = pd.DataFrame(
            [[key, value[0], value[1], value[2], value[3], value[4]]
             for key, value in edges_dict.items()],
            columns=['EDGE_ID', 'TRACK_ID', 'SPOT_SOURCE_ID', 'SPOT_TARGET_ID',
                     'SOURCE_FRAME', 'TARGET_FRAME']
        )
        ordered_edges_df = edges_df.sort_values(
            ['TRACK_ID', 'SOURCE_FRAME', 'TARGET_FRAME'])
        ordered_edges_df.reset_index(inplace=True)

        # compute the differences across successive rows in frame,
        # spot_source and spot_target
        ordered_edges_df['FRAME_DIFF'] = ordered_edges_df.groupby(
            'TRACK_ID')['SOURCE_FRAME'].diff()
        ordered_edges_df['SPOT_SOURCE_DIFF'] = ordered_edges_df.groupby(
            'TRACK_ID')['SPOT_SOURCE_ID'].diff()
        ordered_edges_df['SPOT_TARGET_DIFF'] = ordered_edges_df.groupby(
            'TRACK_ID')['SPOT_TARGET_ID'].diff()

        # create the 'EVENT' column
        ordered_edges_df['EVENT'] = ['None'] * len(ordered_edges_df)
        for i in range(len(ordered_edges_df)):
            tmp = ordered_edges_df.iloc[i]
            # if frame diff is 1: no events
            if tmp['FRAME_DIFF'] == 0:
                if tmp['SPOT_SOURCE_DIFF'] == 0:
                    ordered_edges_df.loc[i, 'EVENT'] = 'split'
                elif tmp['SPOT_TARGET_DIFF'] == 0:
                    ordered_edges_df.loc[i, 'EVENT'] = 'merge'
            elif tmp['FRAME_DIFF'] > 1:
                ordered_edges_df.loc[i, 'EVENT'] = 'gap'
        # shift the event one row up
        ordered_edges_df.EVENT = ordered_edges_df.EVENT.shift(-1)
        return ordered_edges_df

    def read_links(self, ordered_edges_df):
        links_dict = {}
        link_id = 0
        for track in ordered_edges_df.TRACK_ID.unique():
            event = False
            tmp = ordered_edges_df[
                ordered_edges_df.TRACK_ID == track
            ].reset_index()
            links_dict[link_id] = []
            for index, row in tmp.iterrows():
                if row['EVENT'] == 'None' and event is False:
                    links_dict[link_id].append(row['SPOT_SOURCE_ID'])
                    links_dict[link_id].append(row['SPOT_TARGET_ID'])
                    # if source at row zero is not the same as target at row 1,
                    # flag an event
                    if tmp.shape[0] > 1:  # if number rows > 1
                        if (index == 0 and tmp.iloc[index].SPOT_TARGET_ID !=
                                tmp.iloc[index + 1].SPOT_SOURCE_ID):
                            link_id += 1
                            links_dict[link_id] = []
                            event = True
                elif row['EVENT'] == 'split':
                    event = True
                    link_id += 1
                    links_dict[link_id] = []
                    links_dict[link_id].append(row['SPOT_SOURCE_ID'])
                    links_dict[link_id].append(row['SPOT_TARGET_ID'])
                elif row['EVENT'] == 'merge':
                    event = True
                    for key, val in links_dict.items():
                        if row['SPOT_SOURCE_ID'] == val[-1]:
                            links_dict[key].append(row['SPOT_TARGET_ID'])
                            links_dict[key].append(row['SPOT_SOURCE_ID'])
                elif row['EVENT'] == 'gap':
                    if event is False:
                        links_dict[link_id].append(row['SPOT_SOURCE_ID'])
                        links_dict[link_id].append(row['SPOT_TARGET_ID'])
                    elif event is True:
                        for key, val in links_dict.items():
                            if row['SPOT_SOURCE_ID'] == val[-1]:
                                links_dict[key].append(row['SPOT_TARGET_ID'])
                                links_dict[key].append(row['SPOT_SOURCE_ID'])
                    link_id += 1
                    links_dict[link_id] = []
                elif row['EVENT'] == 'None' and event is True:
                    for key, val in links_dict.items():
                        if not val:
                            links_dict[key].append(row['SPOT_SOURCE_ID'])
                            links_dict[key].append(row['SPOT_TARGET_ID'])
                        if row['SPOT_SOURCE_ID'] == val[-1]:
                            links_dict[key].append(row['SPOT_SOURCE_ID'])
                            links_dict[key].append(row['SPOT_TARGET_ID'])
            link_id += 1

        # get only unique spots
        links_dict_unique = {}
        for key, value in links_dict.items():
            unique_set = set(value)
            links_dict_unique[key] = unique_set
        self.logger.debug('Created %d links', len(links_dict_unique))
        links_df = pd.DataFrame()
        for key, value in links_dict_unique.items():
            for spot in value:
                links_df = links_df.append([[key, spot]], ignore_index=True)
        links_df.columns = [cmso.LINK_ID, cmso.OBJECT_ID]
        return links_df


class CellProfilerReader(AbstractReader):

    def read(self):
        self.x = self.conf[config.TRACKING].get(cmso.X_COORD)
        self.y = self.conf[config.TRACKING].get(cmso.Y_COORD)
        self.frame = self.conf[config.TRACKING].get(cmso.FRAME_ID)
        self.obj_id = self.conf[config.TRACKING].get(cmso.OBJECT_ID)
        # parse the digits used for the tracking settings (e.g. 15)
        digits = self.x.split('_')[2]
        self.track_id = 'TrackObjects_Label_' + digits
        self.parent_obj_id = 'TrackObjects_ParentObjectNumber_' + digits
        self.parent_img_id = 'TrackObjects_ParentImageNumber_' + digits
        cp_df, self._objects = self.read_objects()
        self._links = self.read_links(cp_df)

    def read_objects(self):
        objects_dict = {}
        cp_df = pd.read_csv(self.fname)
        cp_df = cp_df.sort_values([self.track_id, self.frame])
        # create new object identifiers
        cp_df.reset_index(inplace=True)
        for index, row in cp_df.iterrows():
            objects_dict[index] = [row[self.frame], row[self.x], row[self.y]]
        objects_df = pd.DataFrame(
            [[key, value[0], value[1], value[2]]
             for key, value in objects_dict.items()],
            columns=[self.obj_id, self.frame, self.x, self.y]
        )
        objects_df.columns = [
            cmso.OBJECT_ID, cmso.FRAME_ID, cmso.X_COORD, cmso.Y_COORD
        ]
        return cp_df, objects_df

    def read_links(self, cp_df):
        links_dict = {}
        LINK_ID = 0
        unique_parent_object = 0
        for track in cp_df[self.track_id].unique():
            tmp = cp_df[cp_df[self.track_id] == track]
            for index, row in tmp.iterrows():
                if index == 0:
                    links_dict[LINK_ID] = [index]
                else:
                    parentImage = row[self.parent_img_id]
                    parentObject = row[self.parent_obj_id]
                    for j, r in tmp.iterrows():
                        if (r.ObjectNumber == parentObject and
                                r[self.frame] == parentImage):
                            unique_parent_object = j
                            break
                    if row.ObjectNumber == row[self.parent_obj_id]:
                        for key, val in links_dict.items():
                            if unique_parent_object == val[-1]:
                                links_dict[key].append(index)
                                break
                    else:
                        LINK_ID += 1
                        links_dict[LINK_ID] = []
                        if row[self.parent_obj_id] != 0:
                            links_dict[LINK_ID].append(unique_parent_object)
                        links_dict[LINK_ID].append(index)
        links_df = pd.DataFrame()
        for key, value in links_dict.items():
            for object_ in value:
                links_df = links_df.append([[key, object_]])
        links_df.columns = [cmso.LINK_ID, cmso.OBJECT_ID]
        return links_df


class IcyReader(AbstractReader):

    def read(self):
        book = xlrd.open_workbook(self.fname)
        sheet = book.sheet_by_index(0)
        track = None
        obj = 0
        objects, links = [], []
        for i in range(sheet.nrows):
            values = sheet.row_values(i)
            if values[0]:  # track number line
                track = int(values[1])
            elif type(values[2]) is float:  # data line
                objects.append([obj] + values[2:6])
                links.append([track, obj])
                obj += 1
        self._objects = pd.DataFrame(
            objects, columns=['OBJECT_ID', 't', 'x', 'y', 'z']
        )
        self._objects.columns = [cmso.OBJECT_ID, cmso.FRAME_ID, cmso.X_COORD,
                                 cmso.Y_COORD, cmso.Z_COORD]
        self._links = pd.DataFrame(links, columns=['LINK_ID', 'OBJECT_ID'])
        self._links.columns = [cmso.LINK_ID, cmso.OBJECT_ID]


class CellmiaReader(AbstractReader):

    ENCODING = "iso-8859-1"
    SEP = "\t"
    CELLMIA_COLS = ["ID of track", "time index", "cell row", "cell col"]
    CMSO_COLS = [cmso.LINK_ID, cmso.FRAME_ID, cmso.Y_COORD, cmso.X_COORD]

    def read(self):
        df = pd.read_csv(self.fname, sep=self.SEP, encoding=self.ENCODING,
                         usecols=self.CELLMIA_COLS)
        df.reset_index(inplace=True)
        cols = [cmso.OBJECT_ID] + self.CMSO_COLS
        df.columns = cols
        cols[-1], cols[-2] = cols[-2], cols[-1]
        df = df[cols]
        self._objects = df.drop(cmso.LINK_ID, 1)
        self._links = df.drop([cmso.FRAME_ID, cmso.X_COORD, cmso.Y_COORD], 1)
        self._links[cmso.LINK_ID] -= 1  # CELLMIA index is 1-based


class MosaicReader(AbstractReader):

    MOSAIC_COLS = ["Trajectory", "Frame", "x", "y", "z"]

    def read(self):
        mo_df = pd.read_csv(self.fname, sep=None, usecols=self.MOSAIC_COLS)
        mo_df.reset_index(inplace=True)
        cols = [cmso.OBJECT_ID, cmso.LINK_ID, cmso.FRAME_ID,
                cmso.X_COORD, cmso.Y_COORD, cmso.Z_COORD]
        mo_df.columns = cols
        self._objects = mo_df.drop(cmso.LINK_ID, 1)
        self._links = mo_df.drop(
            [cmso.FRAME_ID, cmso.X_COORD, cmso.Y_COORD, cmso.Z_COORD], 1)
        self._links[cmso.LINK_ID] -= 1


class TracksReader(object):
    """\
    Generic reader that delegates to specific ones based on file extension.
    """

    def __init__(self, fname, conf=None, log_level=None):
        """\
        Initialize and store a specific reader based on filename extension.

        All other methods delegate to the specific reader.
        """
        logger = get_logger(self.__class__.__name__, level=log_level)
        _, ext = os.path.splitext(fname)
        if ext == '.xls':
            self.reader = IcyReader(fname, conf=conf, log_level=log_level)
        elif ext == '.xml':
            self.reader = TrackMateReader(
                fname, conf=conf, log_level=log_level
            )
        elif ext in ['.csv', '.tsv', '.txt']:
            # read header and delegate to correct reader
            # this pandas method can read csv, tsv and txt files
            df = pd.read_csv(fname, header=None, sep=None, nrows=1)
            trackname = df.iloc[0, 0]
            if trackname == 'ImageNumber':
                self.reader = CellProfilerReader(
                    fname, conf=conf, log_level=log_level
                )
            elif trackname == 'ID of track':
                self.reader = CellmiaReader(
                    fname, conf=conf, log_level=log_level
                )
            elif trackname == 'Trajectory':
                self.reader = MosaicReader(
                    fname, conf=conf, log_level=log_level
                )

        elif ext == '.json':
            self.reader = BiotracksReader(
                fname, conf=conf, log_level=log_level
            )
        else:
            msg = '%r: unknown format: %r' % (fname, ext)
            logger.error(msg)
            raise RuntimeError(msg)

    def read(self):
        self.reader.read()

    @property
    def conf(self):
        return self.reader.conf

    @property
    def objects(self):
        return self.reader.objects

    @property
    def links(self):
        return self.reader.links



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import json
import csv
from copy import deepcopy

import pytest
from datapackage.exceptions import ValidationError

from biotracks import validation, cmso, config

OBJECTS_PATH = "objects.csv"
LINKS_PATH = "links.csv"
TRACKS_PATH = "tracks.csv"
JSON = {
    "name": config.DEFAULT_NAME,
    "resources": [
        {
            "name": cmso.OBJECTS_TABLE,
            "path": OBJECTS_PATH,
            "schema": {
                "fields": [
                    {"name": cmso.OBJECT_ID,
                     "constraints": {"unique": True}},
                    {"name": cmso.FRAME_ID},
                    {"name": cmso.X_COORD},
                    {"name": cmso.Y_COORD},
                ],
                "primaryKey": cmso.OBJECT_ID,
            }
        },
        {
            "name": cmso.LINKS_TABLE,
            "path": LINKS_PATH,
            "schema": {
                "fields": [{"name": cmso.LINK_ID},
                           {"name": cmso.OBJECT_ID}],
                "foreignKeys": [
                    {"fields": cmso.OBJECT_ID,
                     "reference": {"fields": cmso.OBJECT_ID,
                                   "resource": cmso.OBJECTS_TABLE}}
                ]
            }
        }
    ]
}

CSV = {
    OBJECTS_PATH: [
        [cmso.OBJECT_ID, cmso.FRAME_ID,
         cmso.X_COORD, cmso.Y_COORD],
        [0, 1, 0.4, 0.5],
        [1, 2, 0.5, 0.6],
    ],
    LINKS_PATH: [
        [cmso.LINK_ID, cmso.OBJECT_ID],
        [0, 0],
        [0, 1],
    ],
}


@pytest.fixture()
def dp(tmpdir):
    def make_dp(descriptor):
        json_path = tmpdir.join("dp.json")
        with json_path.open("w") as f:
            json.dump(descriptor, f)
        for name, content in CSV.items():
            with tmpdir.join(name).open("w") as f:
                csv.writer(f).writerows(content)
        return str(json_path)
    yield make_dp
    tmpdir.remove(rec=True)


class TestValidation(object):

    def test_validation(self, dp):
        validation.Validator().validate(dp(JSON))

    def test_no_name(self, dp):
        d = deepcopy(JSON)
        del d["name"]
        self.__assert_raises(dp(d))

    def test_no_resources(self, dp):
        d = deepcopy(JSON)
        del d["resources"]
        self.__assert_raises(dp(d))

    def test_bad_name(self, dp):
        d = deepcopy(JSON)
        d["name"] = d["name"].upper()
        self.__assert_raises(dp(d))

    def test_missing_resource(self, dp):
        for i in range(len(JSON["resources"])):
            d = deepcopy(JSON)
            del d["resources"][i]
            self.__assert_raises(dp(d))

    def test_missing_field(self, dp):
        for i, res in enumerate(JSON["resources"]):
            for j in range(len(res["schema"]["fields"])):
                d = deepcopy(JSON)
                del d["resources"][i]["schema"]["fields"][j]
                self.__assert_raises(dp(d))

    def test_primary_key(self, dp):
        obj_i, obj_res = self.__res_by_name(cmso.OBJECTS_TABLE)
        d = deepcopy(JSON)
        del d["resources"][obj_i]["schema"]["primaryKey"]
        self.__assert_raises(dp(d))
        d = deepcopy(JSON)
        d["resources"][obj_i]["schema"]["primaryKey"] = "foobar"
        self.__assert_raises(dp(d))

    def test_constraints(self, dp):
        i, r = self.__res_by_name(cmso.OBJECTS_TABLE)
        j = [_ for (_, f) in enumerate(r["schema"]["fields"])
             if f["name"] == cmso.OBJECT_ID][0]
        d = deepcopy(JSON)
        del d["resources"][i]["schema"]["fields"][j]["constraints"]
        self.__assert_raises(dp(d))
        d = deepcopy(JSON)
        d["resources"][i]["schema"]["fields"][j]["constraints"] = {}
        self.__assert_raises(dp(d))
        d = deepcopy(JSON)
        d["resources"][i]["schema"]["fields"][j]["constraints"]["unique"] = 0
        self.__assert_raises(dp(d))

    def test_foreign_keys(self, dp):
        i, r = self.__res_by_name(cmso.LINKS_TABLE)
        d = deepcopy(JSON)
        del d["resources"][i]["schema"]["foreignKeys"]
        self.__assert_raises(dp(d))
        d = deepcopy(JSON)
        d["resources"][i]["schema"]["foreignKeys"][0]["fields"] = ""
        self.__assert_raises(dp(d))
        for k in "fields", "resource":
            d = deepcopy(JSON)
            d["resources"][i]["schema"]["foreignKeys"][0]["reference"][k] = ""
            self.__assert_raises(dp(d))

    def test_tracks(self, dp):
        tracks_res = {
            "name": cmso.TRACKS_TABLE,
            "path": TRACKS_PATH,
            "schema": {
                "fields": [
                    {"name": cmso.TRACK_ID},
                    {"name": cmso.LINK_ID},
                ],
            },
        }
        d = deepcopy(JSON)
        d["resources"].append(tracks_res)
        validation.Validator().validate(dp(d))
        for i in range(len(tracks_res["schema"]["fields"])):
            d = deepcopy(JSON)
            r = deepcopy(tracks_res)
            del r["schema"]["fields"][i]
            d["resources"].append(r)
            self.__assert_raises(dp(d))

    def __res_by_name(self, name):
        return [(i, r) for (i, r) in enumerate(JSON["resources"])
                if r["name"] == name][0]

    def __assert_raises(self, dp_fn):
        with pytest.raises(ValidationError):
            validation.Validator().validate(dp_fn)



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import math

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def plotXY(df, id_, x_coord, y_coord):
    """Plot the raw trajectories.

    df -- the trajectories dataframe
    id_ -- an identifier (for linear connections of objects)
    x_coord -- the x coordinate
    y_coord -- the y coordinate
    """
    grid = sns.FacetGrid(df, hue=id_,
                         size=6, palette=sns.color_palette("Set1", 10))
    grid.fig.suptitle('XY trajectories')
    grid.map(plt.plot, x_coord, y_coord, marker=".", ms=0.3)
    grid.map(plt.scatter, x_coord, y_coord, marker="o", s=20)
    grid.set_xticklabels(rotation=90)
    sns.plt.savefig("trajectories.png")


def normalize(df, id_, x_coord, y_coord):
    """Normalize to the origin of the coordinate system.

    df -- the trajectories dataframe
    id_ -- an identifier (linkID or trackID)
    x_coord -- the x coordinate
    y_coord -- the y coordinate
    """
    list_ = []
    df = df.dropna()
    for i in df[id_].unique():
        tmp = df[df[id_] == i]
        # convert coordinates columns into numpy array
        array = tmp[[x_coord, y_coord]].values
        # substract first x and y coordinates
        diff_array = array - array[0]
        diff_df = pd.DataFrame(diff_array, columns=['x_norm', 'y_norm'])
        tmp = tmp.assign(x_norm=diff_df.x_norm.values,
                         y_norm=diff_df.y_norm.values)
        list_.append(tmp)
    result = pd.concat(list_)
    return result


def compute_cumulative_displacements(df, id_, x_coord, y_coord):
    """Compute cumulative displacements of motion in the two directions, x
    and y.

    df -- the trajectories dataframe
    id_ -- an identifier
    x_coord -- the x coordinate
    y_coord -- the y coordinate
    """
    list_ = []

    df = df.dropna()
    for i in df[id_].unique():
        tmp = df[df[id_] == i]
        # convert coordinates columns into numpy array
        array = tmp[[x_coord, y_coord]].values
        # add rows
        sum_array = np.cumsum(array, axis=0)
        sum_df = pd.DataFrame(sum_array, columns=['x_cum', 'y_cum'])
        tmp = tmp.assign(x_cum=sum_df.x_cum.values,
                         y_cum=sum_df.y_cum.values)
        list_.append(tmp)

    result = pd.concat(list_)
    return result


def compute_displacements(df, id_, x_coord, y_coord):
    """Compute net displacements of motion in the two directions, x and y.

    df -- the trajectories dataframe
    id_ -- an identifier
    x_coord -- the x coordinate
    y_coord -- the y coordinate
    """
    list_ = []
    df = df.dropna()
    for i in df[id_].unique():
        tmp = df[df[id_] == i]
        # convert coordinates columns into numpy array
        array = tmp[[x_coord, y_coord]].values
        # substract rows
        diff_array = np.diff(array, axis=0)
        # need to insert NaN at the first position
        diff_array = np.insert(diff_array, [0], [np.NaN, np.NaN], axis=0)
        diff_df = pd.DataFrame(diff_array, columns=['delta_x', 'delta_y'])
        tmp = tmp.assign(delta_x=diff_df.delta_x.values,
                         delta_y=diff_df.delta_y.values)
        list_.append(tmp)
    result = pd.concat(list_)
    return result


def compute_turning_angle(df, id_):
    """Compute turning angles.

    df -- the trajectories dataframe
    id_ -- an identifier
    """
    list_ = []
    for i in df[id_].unique():
        tmp = df[df[id_] == i]
        array = tmp[['delta_x', 'delta_y']].values
        turning_angle = np.apply_along_axis(
            lambda x: math.atan2(x[0], x[1]), 1, array)

        ta_df = pd.DataFrame(turning_angle, columns=['ta'])
        tmp = tmp.assign(ta=ta_df.ta.values)
        list_.append(tmp)
    result = pd.concat(list_)
    return result


def plot_polar(theta, N):
    """Plot polar plot.

    theta -- the dataframe with the turning angles
    N -- number of bins to use
    """
    hist, bins = np.histogram(theta.ta, bins=N)

    # the width of the bins interval
    width = [t - s for s, t in zip(bins, bins[1:])]
    bins_ = bins[0:N]  # exclude the last value

    # the actual plotting logic
    g = sns.FacetGrid(theta, size=4)

    radii = hist / max(hist)

    for ax in g.axes.flat:
        ax2 = plt.subplot(111, projection='polar')
        bars = ax2.bar(bins_, radii, width, bottom=0.0)
        for r, bar in zip(radii, bars):
            bar.set_facecolor(plt.cm.Spectral(r))
            bar.set_alpha(0.5)

    sns.plt.savefig("turning_angles.png")



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import os

import pandas as pd
import pytest

from biotracks import cmso, config
from biotracks.pushtopandas import push_to_pandas
from .common import (
    EXAMPLES_DIR, RELPATHS, get_obj_dict, get_link_dict, get_track_dict
)


@pytest.fixture()
def data():
    def make_data(fmt):
        base_dir = os.path.join(EXAMPLES_DIR, fmt, *RELPATHS[fmt][:-1])
        dp_dir = os.path.join(base_dir, 'dp')
        d = {'dp_dir': dp_dir}
        d['obj_df'] = pd.read_csv(os.path.join(dp_dir, 'objects.csv'))
        d['links_df'] = pd.read_csv(os.path.join(dp_dir, 'links.csv'))
        d['tracks_df'] = pd.read_csv(os.path.join(dp_dir, 'tracks.csv'))
        conf_fn = os.path.join(base_dir, config.RELPATH)
        d['conf'] = config.get_conf(conf_fn=conf_fn)
        return d
    return make_data


class TestPushToPandas(object):

    def test_icy(self, data):
        self.__check_dicts(data('ICY'))

    def test_cellmia(self, data):
        self.__check_dicts(data('CELLMIA'))

    def test_cellprofiler(self, data):
        self.__check_dicts(data('CellProfiler'))

    def test_trackmate(self, data):
        self.__check_dicts(data('TrackMate'))

    def test_mosaic(self, data):
        self.__check_dicts(data('Mosaic'))

    def __check_dicts(self, d):
        obj_id = cmso.OBJECT_ID
        link_id = cmso.LINK_ID
        track_id = cmso.TRACK_ID
        ret = push_to_pandas(d['dp_dir'], obj_id)
        exp_track_dict = get_track_dict(d['tracks_df'], link_id, track_id)
        track_dict = get_track_dict(ret['tracks'], link_id, track_id)
        assert track_dict == exp_track_dict
        exp_link_dict = get_link_dict(d['links_df'], obj_id, link_id)
        link_dict = get_link_dict(ret['links'], obj_id, link_id)
        assert link_dict == exp_link_dict
        exp_obj_dict = get_obj_dict(d['obj_df'], obj_id)
        obj_dict = get_obj_dict(ret['objects'], obj_id)
        assert obj_dict.keys() == exp_obj_dict.keys()
        for k, v in exp_obj_dict.items():
            assert obj_dict[k] == pytest.approx(v)



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

"""\
Define standard property names used in CMSO data packages.
"""

OBJECTS_TABLE = "cmso_objects_table"
LINKS_TABLE = "cmso_links_table"
TRACKS_TABLE = "cmso_tracks_table"

X_COORD = "cmso_x_coord"
Y_COORD = "cmso_y_coord"
Z_COORD = "cmso_z_coord"
FRAME_ID = "cmso_frame_id"
OBJECT_ID = "cmso_object_id"
LINK_ID = "cmso_link_id"
TRACK_ID = "cmso_track_id"

SPACE_UNIT = "cmso_space_unit"
TIME_UNIT = "cmso_time_unit"



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as BaseBuildPy
from distutils.errors import DistutilsSetupError
from distutils.dep_util import newer

VERSION_FILE = "VERSION"
PY_VERSION_FILE = "biotracks/version.py"


def get_version_string():
    try:
        with open(VERSION_FILE) as f:
            return f.read().strip()
    except IOError:
        raise DistutilsSetupError("failed to read version info")


def write_version():
    if newer(VERSION_FILE, PY_VERSION_FILE):
        with open(PY_VERSION_FILE, "w") as f:
            f.write("# GENERATED BY setup.py\n")
            f.write("version = '%s'\n" % (get_version_string()))


class BuildPy(BaseBuildPy):

    def run(self):
        write_version()
        BaseBuildPy.run(self)


with open('README.rst') as f:
    readme = f.read()


setup(
    name='biotracks',
    version=get_version_string(),
    description='Data package representation for cell migration tracking data',
    long_description=readme,
    author='paola masuzzo',
    author_email='paola.masuzzo@ugent.be',
    url='https://github.com/CellMigStandOrg/biotracks',
    license="BSD",
    cmdclass={"build_py": BuildPy},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
    ],
    install_requires=[
        "datapackage<1.0",
        "jsontableschema",
        "jsontableschema-pandas"],
    packages=find_packages(exclude=('tests', 'docs'))
)



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import datapackage
import datapackage.registry
import datapackage.exceptions

from . import cmso
from .utils import get_logger


REQUIRED_FIELDS = {
    cmso.OBJECTS_TABLE: {
        cmso.OBJECT_ID, cmso.FRAME_ID, cmso.X_COORD, cmso.Y_COORD
    },
    cmso.LINKS_TABLE: {cmso.LINK_ID, cmso.OBJECT_ID},
    cmso.TRACKS_TABLE: {cmso.TRACK_ID, cmso.LINK_ID},
}

FOREIGN_KEYS = [
    {"fields": cmso.OBJECT_ID,
     "reference": {"fields": cmso.OBJECT_ID,
                   "resource": cmso.OBJECTS_TABLE}}
]


def is_tabular(dp):
    if not isinstance(dp, datapackage.DataPackage):
        return False
    reg = datapackage.registry.Registry()
    return dp.schema.to_dict() == reg.get('tabular')


class ValidationError(datapackage.exceptions.ValidationError):
    pass


class Validator(object):

    def __init__(self, log_level=None):
        self.logger = get_logger(self.__class__.__name__, level=log_level)

    def validate(self, dp):
        if isinstance(dp, datapackage.DataPackage) and not is_tabular(dp):
            raise ValueError("data package must be a tabular data package")
        else:
            dp = datapackage.DataPackage(dp, schema="tabular")
        dp.validate()
        self.logger.debug("valid tabular data package")
        if len(dp.resources) < 2:
            self.__error("data package must have at least two resources")
        res_map = dict((_.descriptor['name'], _) for _ in dp.resources)
        try:
            objects = res_map[cmso.OBJECTS_TABLE]
        except KeyError:
            self.__error("objects table not found")
        else:
            self.validate_objects(objects.descriptor)
        try:
            links = res_map[cmso.LINKS_TABLE]
        except KeyError:
            self.__error("links table not found")
        else:
            self.validate_links(links.descriptor)
        try:
            tracks = res_map[cmso.TRACKS_TABLE]
        except KeyError:
            pass
        else:
            self.validate_tracks(tracks.descriptor)
        return dp

    def validate_objects(self, objects):
        try:
            pk = objects["schema"]["primaryKey"]
        except KeyError:
            self.__error("objects table schema has no primary key")
        if pk != cmso.OBJECT_ID:
            self.__error(
                "objects table primary key must be %r" % (cmso.OBJECT_ID,)
            )
        by_name = self.__check_required_fields(objects)
        id_field = by_name[cmso.OBJECT_ID]
        try:
            constraints = id_field["constraints"]
        except KeyError:
            self.__error("object id field has no constraints")
        try:
            unique = constraints["unique"]
        except KeyError:
            self.__error("object id constraints: missing 'unique' property")
        if not unique:
            self.__error("object id constraints: 'unique' property is false")

    def validate_links(self, links):
        try:
            fk = links["schema"]["foreignKeys"]
        except KeyError:
            self.__error("objects table schema has no foreign keys")
        self.validate_foreign_keys(fk)
        self.__check_required_fields(links)

    def validate_tracks(self, tracks):
        self.__check_required_fields(tracks)

    def validate_foreign_keys(self, fk):
        if len(fk) != 1:
            self.__error("links table must have exactly one foreign key")
        fk = fk[0]
        try:
            fields = fk["fields"]
            ref = fk["reference"]
        except KeyError as e:
            self.__error("missing property in foreignKeys: %r" % e.args)
        if fields != cmso.OBJECT_ID:
            self.__error(
                "foreignKeys fields must be %r" % (cmso.OBJECT_ID,)
            )
        try:
            ref_fields = ref["fields"]
            ref_res = ref["resource"]
        except KeyError as e:
            self.__error(
                "missing property in foreignKeys reference: %r" % e.args
            )
        if ref_fields != cmso.OBJECT_ID:
            self.__error(
                "foreignKeys ref fields must be %r" % (cmso.OBJECT_ID,)
            )
        if ref_res != cmso.OBJECTS_TABLE:
            self.__error(
                "foreignKeys ref resource must be %r" % (cmso.OBJECTS_TABLE,)
            )

    def __check_required_fields(self, descriptor):
        required = REQUIRED_FIELDS[descriptor['name']]
        by_name = dict((_['name'], _) for _ in descriptor['schema']['fields'])
        if not required <= set(by_name):
            self.__error(
                "required fields for %s: %r" % (descriptor['name'], required)
            )
        return by_name

    def __error(self, msg):
        self.logger.error(msg)
        raise ValidationError(msg)



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import json
import csv
import os
import re

import datapackage
from jsontableschema import infer
from . import cmso, config
from .utils import get_logger, mkdir_p


# https://specs.frictionlessdata.io/data-package/#metadata
NAME_PATTERN = re.compile(r"^[a-z0-9_.-]+$")
FOREIGN_KEYS = [{
    "fields": cmso.OBJECT_ID,
    "reference": {
        "datapackage": "",
        "resource": cmso.OBJECTS_TABLE,
        "fields": cmso.OBJECT_ID
    }
}]


def infer_from_df(df, **kwargs):
    # df.iterrows does not preserve types
    h = df.head()
    fields = list(df)
    iterrows = ([str(h[_].values[i]) for _ in fields]
                for i in range(h.shape[0]))
    return infer(fields, iterrows, **kwargs)


def to_json(dp):
    return json.dumps(dp.to_dict(), indent=4, sort_keys=True)


def create(reader, out_dir, log_level=None):
    logger = get_logger("createdp.create", level=log_level)
    top_level_dict = reader.conf[config.TOP_LEVEL]
    try:
        name = top_level_dict["name"]
    except KeyError:
        raise ValueError("'name' is a required property")
    if not NAME_PATTERN.match(name):
        raise ValueError("invalid name: %r" % (name,))
    dp = datapackage.DataPackage()
    for k, v in top_level_dict.items():
        dp.descriptor[k] = v
    dp.descriptor['resources'] = []
    mkdir_p(out_dir)
    logger.info("writing to '%s'", out_dir)
    for a in "objects", "links":
        out_bn = "%s.csv" % a
        out_fn = os.path.join(out_dir, out_bn)
        df = getattr(reader, a)
        df.to_csv(out_fn, index=False, quoting=csv.QUOTE_NONE)
        if a == "objects":
            name = cmso.OBJECTS_TABLE
            infer_kwargs = {"primary_key": cmso.OBJECT_ID}
        else:
            name = cmso.LINKS_TABLE
            infer_kwargs = {}
        schema = infer_from_df(df, **infer_kwargs)
        if a == "links":
            schema['foreignKeys'] = FOREIGN_KEYS
        res = {"name": name, "path": out_bn, "schema": schema}
        dp.descriptor['resources'].append(res)
    with open(os.path.join(out_dir, 'dp.json'), 'w') as f:
        f.write(to_json(dp) + '\n')
    return dp



# #%L
# Copyright (c) 2016-2017 Cell Migration Standardisation Organization
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# #L%

import os

import pandas as pd
import pytest

from biotracks import readfile, cmso, config
from .common import EXAMPLES_DIR, RELPATHS, get_obj_dict, get_link_dict


@pytest.fixture()
def data():
    def make_data(fmt):
        base_dir = os.path.join(EXAMPLES_DIR, fmt, *RELPATHS[fmt][:-1])
        in_fn = os.path.join(base_dir, RELPATHS[fmt][-1])
        conf_fn = os.path.join(base_dir, config.RELPATH)
        conf = config.get_conf(conf_fn=conf_fn)
        reader = readfile.TracksReader(in_fn, conf=conf)
        d = {'reader': reader}
        for k in 'objects', 'links':
            d['%s_path' % k] = os.path.join(base_dir, 'dp', '%s.csv' % k)
        dp_path = os.path.join(base_dir, 'dp', 'dp.json')
        d['bt_reader'] = readfile.BiotracksReader(dp_path, conf=conf)
        return d
    return make_data


class TestReadFile(object):

    def test_icy(self, data):
        self.__check_dicts(data('ICY'))

    def test_cellmia(self, data):
        self.__check_dicts(data('CELLMIA'))

    def test_cellprofiler(self, data):
        self.__check_dicts(data('CellProfiler'))

    def test_trackmate(self, data):
        d = data('TrackMate')
        self.__check_dicts(d)
        top_level = d['reader'].conf[config.TOP_LEVEL]
        from_source = {}
        for k in cmso.SPACE_UNIT, cmso.TIME_UNIT:
            assert k in top_level
            from_source[k] = top_level[k]
        # check override
        spatial_unit = from_source[cmso.SPACE_UNIT] + "_"
        top_level[cmso.SPACE_UNIT] = spatial_unit
        d['reader'].read()
        assert top_level[cmso.SPACE_UNIT] == spatial_unit

    def test_mosaic(self, data):
        self.__check_dicts(data('Mosaic'))

    def __check_dicts(self, d):
        for reader in (d['reader'], d['bt_reader']):
            for name in 'objects', 'links':
                assert type(getattr(reader, name) is pd.DataFrame)
            obj_id, link_id = cmso.OBJECT_ID, cmso.LINK_ID
            exp_link_dict = get_link_dict(
                pd.read_csv(d['links_path']), obj_id, link_id
            )
            link_dict = get_link_dict(reader.links, obj_id, link_id)
            assert link_dict == exp_link_dict
            exp_obj_dict = get_obj_dict(pd.read_csv(d['objects_path']), obj_id)
            obj_dict = get_obj_dict(reader.objects, obj_id)
            assert obj_dict.keys() == exp_obj_dict.keys()
            for k, v in exp_obj_dict.items():
                assert obj_dict[k] == pytest.approx(v)
